"""
edx_django_utils Django application initialization.
"""

from django.apps import AppConfig


class EdxDjangoUtilsConfig(AppConfig):
    """
    Configuration for the edx_django_utils Django application.
    """

    name = 'edx_django_utils'
